''' entity classes '''

import random

import es

import srcds.srcds_types as srcds_types
import srcds.srcds as srcds

Vector = srcds_types.Vector
Color = srcds_types.Color
RangedValue = srcds_types.RangedValue
SymbolicValue = srcds_types.SymbolicValue

SRCDSError = srcds.SRCDSError

INVALID_INDEX = -1


class _KeyValue(object):
    ''' Internal class to handle entity keyvalues. '''
    def __init__(self, keyname, keyvalue, readonly, getvalue, autoupdate, applyvalue):
        self._keyname = keyname
        self._keyvalue = keyvalue
        self._readonly = readonly
        self._getvalue = getvalue
        self._autoupdate = autoupdate
        self._applyvalue = applyvalue
        self._type = type(keyvalue)
    
    def setvalue(self, value):
        ''' Sets a new value for key, if allowed. '''
        if self._readonly:
            raise TypeError('Key %s is read-only'%self._keyname)
        
        if hasattr(self._keyvalue, 'setvalue'):
            # a custom type designed for this
            self._keyvalue.setvalue(value)
        elif type(value) != self._type:
            # incompatible type, try creating a new instance of the original type
            try:
                self._keyvalue = self._type(value)
            except ValueError:
                # unknown value, possibly this keyvalue does not exist in the game
                return False
        else:
            # compatible type, so set the value there
            self._keyvalue = value
        
        return self._autoupdate

    def getvalue(self):
        ''' Gets the key value as is. '''
        if hasattr(self._keyvalue, 'getvalue'):
            # a custom type that knows what to return here
            return self._keyvalue.getvalue()
        return self._keyvalue

    def __str__(self):
        ''' Returns the key value as string. '''
        return str(self._keyvalue)

    def __int__(self):
        ''' Returns the key value as integer. '''
        return int(self._keyvalue)

    def __float__(self):
        ''' Returns the key value as float. '''
        return float(self._keyvalue)

    def __coerce__(self, other):
        ''' Coerces to the other type. '''
        return type(other)(self._keyvalue)


class BaseEntity(object):
    '''
Base entity class for all entities.

Adds the following keyvalues to entities:
    targetname <target_source>
        Defines the name that other entities refer to this entity by.

    hammerid <integer readonly>
        This id is used for debugging purposes in Hammer.

    classname <readonly>
        The entity's classname (e.g. info_target)

    angles <pitch yaw roll>
        An angular world orientation in degrees.

    origin <x y z>
        The location in the world.

    spawnflags <integer>
        Flags that can be set using x.set_flags, x.clear_flags and x.toggle_flags.
'''
    
    # Keyvalues data defined by this entity type
    # _data = {
    #     'keyname': default_value,
    #     'keyname': default_value,
    #     ...
    # }
    _data = {
        'targetname': '',
        'angles': Vector((0,0,0)),
        'origin': Vector((0,0,0)),
    }
    
    def __init__(self, index=INVALID_INDEX):
        attributes = vars(self)
        attributes['_attributes'] = attributes #hax
        attributes['_keyvalues'] = {}
        self._add('classname', self.__class__.__name__, readonly=True, getvalue=False, autoupdate=False, applyvalue=False)

        if index == INVALID_INDEX:
            self._create()
        else:
            self._index = index
        
        self._add('spawnflags', 0, autoupdate=False)
        #self._add('hammerid', 0, readonly=True, applyvalue=False)
        cls = self.__class__
        for entclass in reversed(cls.__mro__):
            if hasattr(entclass, '_data'):
                for key, value in entclass._data.iteritems():
                    self._add(key, value)

    def __str__(self):
        ''' str(x), returns the targetname of the entity. '''
        return str(self.targetname)

    def __int__(self):
        ''' int(x), returns the entity index of the entity. '''
        return int(self._index)

    def __setattr__(self, attribute, value):
        ''' x.attribute = value '''
        if attribute[0] != '_' and attribute in self._keyvalues:
            es.dbgmsg(1, 'Using kvinst')
            kvinst = self._keyvalues[attribute]
            if kvinst.setvalue(value):
                es.entitysetvalue(self._index, attribute, kvinst.getvalue())
        elif attribute[0] == '_':
            self._attributes[attribute] = value
        else:
            raise AttributeError("'%s' entity has no attribute '%s'"%(
                    self.classname, attribute
                    )
            )
    
    def __getattr__(self, attribute):
        ''' y = x.attribute '''
        if attribute[0] != '_' and attribute in self._keyvalues:
            return self._keyvalues[attribute].getvalue()
        else:
            try:
                return self._attributes[attribute]
            except KeyError:
                raise AttributeError("'%s' entity has no attribute '%s'"%(
                    self.classname, attribute
                    )
                )
        
    def _add(self, keyname, keyvalue, readonly=False, getvalue=True, autoupdate=True, applyvalue=True):
        ''' Internal method to add keyvalues. '''
        self._keyvalues[keyname] = _KeyValue(keyname, keyvalue, readonly, getvalue, autoupdate, applyvalue)

    def _getvalues(self, index):
        ''' Fills in keyvalues from an existing entity. '''
        for attribute in self._keyvalues:
            es.dbgmsg(1, 'Getting attribute %s for entity %d'%(attribute, index))
            if self._keyvalues[attribute]._getvalue:
                value = es.entitygetvalue(index, attribute)
                es.dbgmsg(1, 'value = %s'%repr(value))
                self.__setattr__(attribute, value)

    def set_flags(self, *flags):
        ''' Sets spawnflags. '''
        for flag in flags:
            if isinstance(flag, int):
                nflag = int(flag)
            else:
                nflag = self._flags[flag]
            self.spawnflags |= nflag

    def toggle_flags(self, *flags):
        ''' Toggles spawnflags. '''
        for flag in flags:
            if isinstance(flag, int):
                nflag = int(flag)
            else:
                nflag = self._flags[flag]
            self.spawnflags ^= nflag

    def clear_flags(self, *flags):
        ''' Clears spawnflags. '''
        for flag in flags:
            if isinstance(flag, int):
                nflag = ~int(flag)
            else:
                nflag = ~self._flags[flag]
            self.spawnflags &= nflag

    def Kill(self):
        ''' Removes this entity from the world. '''
        self.fire('Kill')

    def KillHierarchy(self):
        ''' Removes this entity and all its children from the world. '''
        self.fire('KillHierarchy')

    def AddOutput(self, text):
        '''
Evaluates a keyvalue/output on this entity.
Format: <key> <value>
Format: <output name> <targetname>,<inputname>,<parameter>,<delay>,<max times to fire (-1 == infinite)>
        '''
        keyname, keyvalue = text.split(' ', 1)
        if keyname in self._keyvalues:
            self.__setattr__(keyname, keyvalue)
        
        self.fire('AddOutput "%s"'%text)

    def FireUser1(self):
        ''' Causes this entity's OnUser1 output to be fired. '''
        self.FireUser(self, 1)

    def FireUser2(self):
        ''' Causes this entity's OnUser2 output to be fired. '''
        self.FireUser(self, 2)

    def FireUser3(self):
        ''' Causes this entity's OnUser3 output to be fired. '''
        self.FireUser(self, 3)

    def FireUser4(self):
        ''' Causes this entity's OnUser4 output to be fired. '''
        self.FireUser(self, 4)

    def FireUser(self, num):
        ''' Causes this entity's OnUser<num> output to be fired. '''
        self.fire('FireUser%d'%num)

    def _create(self):
        ''' Creates the entity and assigns it entity index '''
        self._index = es.createentity(self.classname)
        return self._index
    
    def spawn(self):
        ''' Spawns this entity to the world. '''
        self._set_spawn_flags()
        es.spawnentity(self._index)

    def _set_spawn_flags(self):
        ''' Applies the spawnflags. '''
        es.entitysetvalue(self._index, 'spawnflags', int(self.spawnflags))

    def _random_name(self):
        ''' Generates a random targetname. '''
        return "%s_%d"%(self.__class__.__name__, int(random.random()*1000000))

    def fire(self, fire_input):
        ''' Fires an input on the entity. '''
        userid = es.getuserid()
        if userid:
            randomname = self._random_name()
            es.setentityname(self._index, randomname)
            es.fire(userid, randomname, fire_input)
            es.setentityname(self._index, self.targetname)
        else:
            raise SRCDSError('''Cannot fire inputs when no users are active.
        Your input was <%s>'''%fire_input)

    def fire_group(self, fire_input):
        ''' Fires an input on all the entities that share the targetname/classname. '''
        userid = es.getuserid()
        if userid:
            es.fire(userid, self.targetname or self.classname, fire_input)
        else:
            raise SRCDSError('''Cannot fire inputs when no users are active.
        Your input was <%s>'''%fire_input)

    def copy(self):
        ''' Creates a new entity using this entity as a template. '''
        ent_class = type(self)
        if ent_class is UnknownEntity:
            raise SRCDSError('Cannot copy unknown entities')
            new_entity = ent_class(index=self._index)
            new_entity.classname = self.classname
            new_entity._create()
        else:
            new_entity = ent_class()
        new_entity._keyvalues = self._keyvalues.copy()
        return new_entity


class UnknownEntity(BaseEntity):
    '''
Entity class used for entities not explicitly implemented.
    '''
    def __init__(self, *args, **kw):
        super(UnknownEntity, self).__init__(*args, **kw)
        self._add('classname', 'unknown', readonly=False, getvalue=True, autoupdate=False, applyvalue=False)
    
    def __setattr__(self, attribute, value):
        ''' x.attribute = value '''
        if attribute[0] != '_':
            if attribute in self._keyvalues:
                es.dbgmsg(1, 'Using kvinst')
                kvinst = self._keyvalues[attribute]
                if kvinst.setvalue(value):
                    es.entitysetvalue(self._index, attribute, kvinst.getvalue())
            else:
                es.dbgmsg(1, 'Using direct setvalue')
                es.entitysetvalue(self._index, attribute, value)
        else:
            self._attributes[attribute] = value


class BaseParentableEntity(BaseEntity):
    '''
Base entity class for parentable entities.

Adds the following keyvalues to entities:
    parentname <target_destination>
        Specifies the targetname of this entity's Parent in a movement hierarchy. Child entities mimic the movements of their Parent.
'''
    
    # Keyvalues data defined by this entity type
    # _data = {
    #     'keyname': default_value,
    #     'keyname': default_value,
    #     ...
    # }
    _data = {
        'parentname': '',
    }

    def SetParent(self, target):
        ''' Changes the entity's parent in the movement hierarchy. '''
        self.parentname = target
        self.fire('SetParent %s'%target)

    def SetParentAttachment(self, attachment):
        '''
Change this entity to attach to a specific attachment point on its parent.
Entities must be parented before being sent this input.
The parameter passed in should be the name of the attachment.
        '''
        self.fire('SetParentAttachment %s'%attachment)

    def SetParentAttachmentMaintainOffset(self, attachment):
        '''
Change this entity to attach to a specific attachment point on it's parent.
Entities must be parented before being sent this input.
The parameter passed in should be the name of the attachment.
The entity will maintain it's position relative to the Parent's Origin and Angles at the time it is attached.
        '''
        self.fire('SetParentAttachmentMaintainOffset %s'%attachment)

    def ClearParent(self):
        ''' Removes this entity from its current movement hierarchy. '''
        self.fire('ClearParent')


class BaseRenderedEntity(BaseEntity):
    '''
Base entity class for rendered entities.

Adds the following keyvalues to entities:
    renderfx <choices>
        See object.renderfx_values for the list of applicable rendering effects.

    HDRColorScale <float>
        Float value to multiply sprite color by when running in HDR mode.

    renderamt <byte>
        The strength of the rendering effect. (0-255)

    rendercolor <color255>
        Rendering color.
'''
    
    renderfx_values = {
        'Normal': 0,
        'Slow Pulse': 1,
        'Fast Pulse': 2,
        'Slow Wide Pulse': 3,
        'Fast Wide Pulse': 4,
        'Slow Fade Away': 5,
        'Fast Fade Away': 6,
        'Slow Become Solid': 7,
        'Fast Become Solid': 8,
        'Slow Strobe': 9,
        'Fast Strobe': 10,
        'Faster Strobe': 11,
        'Slow Flicker': 12,
        'Fast Flicker': 13,
        'Constant Glow': 14,
        'Distort': 15,
        'Hologram': 16,
        'Scale up': 17,
        'Spotlight FX': 22,
        'Cull By Distance': 23,
        'Fade Wider Pulse': 24,
    }

    # Keyvalues data defined by this entity type
    # _data = {
    #     'keyname': default_value,
    #     'keyname': default_value,
    #     ...
    # }
    _data = {
        'renderfx': SymbolicValue('Normal', renderfx_values),
        'HDRColorScale': 1.0,
        'renderamt': RangedValue(0,0,255),
        'rendercolor': Color(0,0,0),
    }

    def Color(self, color):
        ''' Sets the rendercolor '''
        self.rendercolor = color
        self.fire('Color "%s"'%str(self.rendercolor))

    def Alpha(self, alpha):
        ''' Sets the renderamt '''
        self.renderamt = alpha & 0xff
        self.fire('Alpha "%s"'%str(self.renderamt))


class env_beam(BaseParentableEntity, BaseRenderedEntity):
    '''
Creates a visible energy beam running between two entities, in either a line
or ring shape. Similar to env_laser.

For the list of attributes and methods, see the Valve Developer Community wiki
http://developer.valvesoftware.com/wiki/Env_beam
    '''

    # These constants are defined in Source SDK
    MAX_BEAM_WIDTH = 25.5
    MAX_BEAM_NOISEAMPLITUDE = 64.0
    MAX_BEAM_SCROLLSPEED = 100.0

    TouchType_values = {
        'Not a tripwire': 0,
        'Player Only': 1,
        'NPC Only': 2,
        'Player or NPC': 3,
        'Player or NPC or Physprop': 4,
    }

    _flags = {
        'Start On': 1,
        'Toggle': 2,
        'Random Strike': 4,
        'Ring': 8,
        'StartSparks': 16,
        'EndSparks': 32,
        'Decal End': 64,
        'Shade Start': 128,
        'Shade End': 256,
        'Taper Out': 512,
    }

    # Keyvalues data defined by this entity type
    # _data = {
    #     'keyname': default_value,
    #     'keyname': default_value,
    #     ...
    # }
    _data = {
        'Radius': 0,
        'life': 0,
        'BoltWidth': RangedValue(1.0, 1.0, MAX_BEAM_WIDTH),
        'NoiseAmplitude': RangedValue(0.0, 0.0, MAX_BEAM_NOISEAMPLITUDE),
        'texture': '',
        'TextureScroll': RangedValue(0.0, 0.0, MAX_BEAM_SCROLLSPEED),
        'framerate': 0,
        'framestart': 0,
        'StrikeTime': 0.0,
        'damage': 0.0,
        'LightningStart': '',
        'LightningEnd': '',
        'TouchType': SymbolicValue('Not a tripwire', TouchType_values),
        'filtername': '',
        'decalname': '',
    }

    def TurnOn(self):
        ''' Turns the beam on. '''
        self.fire('TurnOn')

    def TurnOff(self):
        ''' Turns the beam off. '''
        self.fire('TurnOff')

    def Toggle(self):
        ''' Toggles the beam's state. '''
        self.fire('Toggle')

    def StrikeOnce(self):
        '''
        Causes the beam to strike once. It will stay on for its set Life and
        then turn off (it will never turn off if Life is set to zero).
        '''
        self.fire('StrikeOnce')

    def Noise(self, amplitude):
        ''' Set the NoiseAmplitude. '''
        self.fire('Noise %s'%amplitude)

    def ScrollSpeed(self, speed):
        ''' Set the TextureScroll. '''
        self.fire('ScrollSpeed %s'%speed)

    def Width(self, width):
        ''' Set the BoltWidth. '''
        self.fire('Width %s'%width)


class info_zombie_spawn(BaseEntity):
    '''
A select spawnpoint for a random infected.

For the list of attributes and methods, see the Valve Developer Community wiki page
http://developer.valvesoftware.com/wiki/Info_zombie_spawn
    '''
    _flags = {
    }

    _data = {
        'Popuplation': '',
    }
    
    def SpawnZombie(self):
        '''Spawns a random infected.'''
        self.fire('SpawnZombie')


class player(BaseParentableEntity, BaseRenderedEntity):
    '''
The player. This entity still exists even if the player is dead.

For the list of attributes and methods, see the Valve Developer Community wiki page
http://developer.valvesoftware.com/wiki/Player
    '''

    _flags = {
    }

    _data = {
        'friction': 1.0,
        'health': 100,
        'max_health': 100,
        'physdamagescale': 1.0,
    }
    
    def IgnoreFallDamage(self):
        ''' IgnoreFallDamage (undocumented) '''
        self.fire('IgnoreFallDamage')

    def IgnoreFallDamageWithoutReset(self):
        ''' IgnoreFallDamage (do not reset) '''
        self.fire('IgnoreFallDamageWithoutReset')
    
    def OnSquadMemberKilled(self, param=None):
        ''' OnSquadMemberKilled (undocumented, shouldn't this be output?) '''
        if param:
            self.fire('OnSquadMemberKilled "%s"'%str(param))
        else:
            self.fire('OnSquadMemberKilled')
    
    def DisableFlashlight(self):
        ''' Disables the player's flashlight '''
        self.fire('DisableFlashlight')
    
    def EnableFlashlight(self):
        ''' Enables the player's flashlight '''
        self.fire('EnableFlashlight')
    
    def ForceDropPhysObjects(self):
        ''' Forces the player drop any grabbed physics objects (undocumented) '''
        self.fire('ForceDropPhysObjects')
    
    def SetHealth(self, health):
        ''' Sets player health '''
        self.health = health
        self.fire('Health "%s"'%str(self.health))
    
    def SetHUDVisibility(self, value):
        ''' Sets HUD visibility (0 = off, 1 = on) '''
        self.fire('SetHUDVisibility "%s"'%str(value))
    
    def SetFogController(self, controller):
        ''' Set Fog Controller for the player'''
        self.fire('SetFogController "%s"'%str(controller))
    
    def KilledNPC(self, param):
        ''' Killed a non-playing character (undocumented) '''
        if param:
            self.fire('KilledNPC "%s"'%str(param))
        else:
            self.fire('KilledNPC')

    def skin(self, skin_name):
        ''' Sets the model's skin '''
        self.fire('skin "%s"'%str(skin_name))


class prop_physics(BaseRenderedEntity):
    '''
    Used to add rigid, VPhysics-simulated models to the world. 

    The only value required for prop_physics to work is a path to a model.
    As long as that model is physics-enabled, the correct data will be loaded
    from it when the map starts. It is possible to change any embedded physics
    data with the "Override Parameters" keyvalue.

    http://developer.valvesoftware.com/wiki/Prop_physics
    '''
    _flags = {
    }

    _data = {
    }

class prop_physics_multiplayer(prop_physics):
    '''
Identical to prop_physics except that it bounces away from collision
(if sv_turbophysics is enabled), which helps to avoid the prediction
errors that normal physics objects typically generate.

See
http://developer.valvesoftware.com/wiki/Prop_physics_multiplayer
    '''
    _flags = {
    }

    _data = {
    }

class weapon_ammo_spawn(BaseEntity):
    '''
A possible spawnpoint for ammo restock in Left 4 Dead.

For the list of attributes and methods, see the Valve Developer Community wiki page
http://developer.valvesoftware.com/wiki/Weapon_ammo_spawn
    '''
    _flags = {
    }

    _data = {
        'count': 0,
        'ammo': 0,
        'model': 'models/props/terror/ammo_stack.mdl',
    }


